from hashlib import blake2b
import xml.etree.ElementTree as ET
import sys

class DwCMetadata:

    def __init__(self, eml):
        self.allLines = []
        self.emls = eml
        self.valid = True

    def convert(self):
        try:
            self.allLines = []
            root = ET.fromstring(self.emls)
            dataset = root.find('./dataset')
            if dataset is None:
                return
            self._title(dataset.find('./title'))
            for ai in dataset.findall('./alternateIdentifier'):
                self._element(ai)
            self._element(dataset.find('./pubDate'))
            self._element(dataset.find('./language'))
            self._person(dataset.find('./creator'))
            self._person(dataset.find('./metadataProvider'))
            self._person(dataset.find('./associatedParty'))

            self._abstract(dataset.find('./abstract'))
            self._keywords(dataset)
            self._intellectualRigths(dataset)
            self._geographicCoverage(dataset.find('./coverage/geographicCoverage'))
            self._taxonomicCoverage(dataset.find('./coverage/taxonomicCoverage'))

            self._maintenance(dataset.find('./maintenance'))
            self._person(dataset.find('./contact'))
            self._methods(dataset.find('./methods'))
            self._project(dataset.find('./project'))
            self._additionalMetadata(root.find('./additionalMetadata'))
            h = blake2b(key=b'FrictionlessDarwinCore', digest_size=16)
            for line in self.allLines:
                h.update(line.encode('utf-8'))
            self._about()
            self.hexdigest = h.hexdigest()
        except BaseException as err:
            self.valid = False
            print(err, 'in DwCMetadata.convert()')
        else:
            self.valid = True
        return

    def as_markdown(self):
        s = '\n\n'
        return s.join(self.allLines)

    def hexdigest(self):
        return self.hexdigest

    def _about(self):
        self._addLine('---')
        self._addLine('generated by [FrictionlessDarwinCore](https://github.com/frictionlessdata/FrictionlessDarwinCore)')

    def _addLine(self, line):
        if line is not None:
            self.allLines.append(line)

    def _title(self, element):
        if element is not None:
            self._addLine('# ' + element.text)

    def _element(self, element):
        if element is not None and element.text is not None:
            self._addLine(element.tag + ': ' + element.text)

    def _abstract(self, element):
        if element is not None:
            self._addLine('## '+ element.tag)
            for p in element.findall('./para'):
                self._addLine(p.text)

    def _keywords(self, element):
        if element is not None:
            self._addLine('## Keywords')
            for k in element.findall('./keywordSet'):
                self._addLine('*'+ k.findtext('./keyword') + '* ' + k.findtext('./keywordThesaurus'))

    def _intellectualRigths(self, element):
        if element is not None:
            ipr = element.find('./intellectualRights')
            if ipr is not None:
                para = ipr.find('./para')
                if para is not None:
                    self._addLine('## Intellectual Property Rights')
                    ulink = para.find('./ulink')
                    if ulink is not None:
                        citetitle = ulink.find('./citetitle')
                        ct = 'see'
                        if citetitle is not None:
                            ct = citetitle.text
                        self._addLine(' [' + ct + '](' + ulink.get('url') + ')')
                    else:
                        self._addLine(para.text)

    def _geographicCoverage(self, element):
        if element is not None:
            self._addLine('## Geographic Coverage')
            gd = element.find('./geographicDescription')
            if gd is not None:
                self._element(gd)
            self._addLine('BoundingCoordinates: West:' + element.findtext('./boundingCoordinates/westBoundingCoordinate') +
                '°, East:' + element.findtext('./boundingCoordinates/eastBoundingCoordinate') +
                '°, North:' + element.findtext('./boundingCoordinates/northBoundingCoordinate') +
                '°, South:' + element.findtext('./boundingCoordinates/southBoundingCoordinate') + '°')

    def _taxonomicCoverage(self, element):
        if element is not None:
            self._addLine('## Taxonomic Coverage')
            gtc=element.find('./generalTaxonomicCoverage')
            if gtc is not None:
                self._element(gtc)
            self._addLine('### Classification')
            for tc in element.findall('./taxonomicClassification/*'):
                self._element(tc)

    def _maintenance(self, element):
        if element is not None:
            self._addLine('## Maintenance')
            self._addLine(element.findtext('./description/para'))
            muf=element.find('./maintenanceUpdateFrequency')
            if muf is not None:
                self._element(muf)

    def _methods(self, element):
        if element is not None:
            self._addLine('## Methods')
            ms=element.findtext('./methodStep/description/para')
            if ms is not None:
                self._addLine('methodStep: ' + ms)
            se=element.findtext('./sampling/studyExtent/description/para')
            if se is not None:
                self._addLine('studyExtent: ' + se)
            sd=element.findtext('./sampling/samplingDescription/para')
            if sd is not None:
                self._addLine('samplingDescription: ' + sd)

    def _project(self, element):
        if element is not None:
            self._addLine('## Project')
            id=element.get('id')
            if id is not None:
                self._addLine('id='+element.get('id'))
            title = element.find('./title')
            if title is not None:
                self._element(title)
            self._person(element.find('./personnel'))
            self._abstract(element.find('./abstract'))
            self._abstract(element.find('./funding'))
            sad=element.findtext('./studyAreaDescription/descriptor/descriptorValue')
            if sad is not None:
                self._addLine('## study area description')
                self._addLine(sad)
            dd=element.findtext('./designDescription/description/para')
            if dd is not None:
                self._addLine('## design description')
                self._addLine(dd)

    def _person(self, element):
        if element is not None:
            self._addLine('## ' + element.tag)
            ind=element.find('./individualName')
            if ind is not None:
                name=''
                if element.find('./givenName') is not None:
                    name = element.find('./givenName') + ' '
                if element.find('./surName') is not None:
                    element.findtext('./surName')
                self._addLine('Name:'+name)
            org=element.findtext('./organizationName')
            if org is not None:
                self._addLine('Organization:'+ org)
            pos=element.findtext('./positionName')
            if pos is not None:
                self._addLine('Position:'+ pos)
            email=element.findtext('./electronicMailAddress')
            if email is not None:
                self._addLine('email:'+email)
            userid=element.find('./userId')
            if userid is not None:
                self._addLine('userId:'+userid.text+ ' ('+ userid.get('directory') + ')')

    def _additionalMetadata(self, element):
        if element is not None:
            self._addLine('## Additional Metadata')
            for am in element.findall('./metadata/gbif/*'):
                self._element(am)




from FrictionlessDarwinCore import DwCVocabulary
import xml.etree.ElementTree as ET
import re
import json


class DwCStructure:
    voc = DwCVocabulary()
    ns = {'dwc': 'http://rs.tdwg.org/dwc/text/'}

    def __init__(self, meta, eml, meta_dir =''):
        self.eml = eml
        self.meta = meta
        self.meta_dir = meta_dir
        self.descriptor = {}
        self.corename = ''
        self.pkey_name = 'id'
        self.has_default_values = False
        self.valid = True

    def convert(self):
        # convert meta.xml into datapackage descriptor
        try:
            self.descriptor = {}
            if self.eml is not None:
                dataset = ET.fromstring(self.eml).find('./dataset')
                self._addheader(dataset)
            if self.meta is not None:
                archive = ET.fromstring(self.meta)
                resources = []
                core = archive.find('dwc:core', DwCStructure.ns)
                self.corename = core.get('rowType').rsplit('/', 1)[1].lower()

                resources.append(self._toresource(core, True))
                for extension in archive.findall('dwc:extension', DwCStructure.ns):
                    resources.append(self._toresource(extension, False))
                self._add('resources', resources)
        except BaseException as err:
            self.valid = False
            print(err, 'in DwCStructure.convert()')
        else:
            self.valid = True
        return self.as_json()

    def as_json(self):
        return json.dumps(self.descriptor)


    def _add(self, key, value):
        self.descriptor[key] = value

    def _addheader(self, dataset):
        licences = []
        sources = []
        for ai in dataset.findall('./alternateIdentifier'):
            m = re.search(r'(?<==)\w+', ai.text)
            if m is not None:
                self._add('name', m.group(0))
            if re.search('^[0-9a-f\-]*$', ai.text):
                self._add('id', ai.text)
            if re.search('^https?://', ai.text):
                source = {'title': 'GBIF IPT', 'path': ai.text}
                sources.append(source)
        title=dataset.find('.title')
        if title is not None:
            self._add('title', title.text)
        self._add('profile', 'tabular-data-package')
        ipr = dataset.find('./intellectualRights')
        if ipr is not None:
            licence = {}
            ulink = ipr.find('./para/ulink')
            if ulink is not None:
                ct = ulink.findtext('./citetitle')
                if ct is not None:
                    name = re.search('\((.+)\)', ct)
                    if name is not None:
                        licence['name'] = name.group(0)
                    licence['title'] = ct
                licence['path'] = ulink.get('url')
            licences.append(licence)
        self._add('licences', licences)

        self._add('sources', sources)

    def _delimiter(self, delimiter_string):
        switcher = {
            "\\n": '\n',
            "\\r": '\r',
            "\\r\\n": '\r\n',
            "\\t": '\t',
        }
        if delimiter_string is not None:
            return switcher.get(delimiter_string, delimiter_string)
        else:
            return ''

    def _toresource(self, mfile, core):
        if mfile is None:
            return
        r = {}
        files = mfile.find('dwc:files', DwCStructure.ns)
        location = files.find('dwc:location', DwCStructure.ns)
        r['name'] = location.text.split('.')[0]
        r['path'] = location.text
        r['description'] = mfile.get('rowType')
        r['profile'] = 'tabular-data-resource'
        r['encoding'] = mfile.get('encoding')
        r['format'] = 'csv'
        dialect = {'csvddfVersion': 1.2, 'delimiter': self._delimiter(mfile.get('fieldsTerminatedBy')),
                   'doubleQuote': True, 'lineTerminator': self._delimiter(mfile.get('linesTerminatedBy')),
                   'quoteChar': self._delimiter(mfile.get('fieldsEnclosedBy')), 'skipInitialSpace': True,
                   'header': mfile.get('ignoreHeaderLines') == '1', 'commentChar': '#'}
        r['dialect'] = dialect

        schema = {}
        fields = []

        index_id='0'
        index_name = 'id'

        if core:
            index = mfile.find('dwc:id', DwCStructure.ns)
        else:
            index = mfile.find('dwc:coreid', DwCStructure.ns)
        if index is not None:
            index_id = index.get('index')

        need_additional_id_field = True
        for f in mfile.findall('dwc:field', DwCStructure.ns):
            if f.get('index') == index_id:
                need_additional_id_field = False
                index_name = f.get('term').rsplit('/', 1)[1]
                if core:
                    self.pkey_name = index_name

        if need_additional_id_field:
            print('Adding additional_id_field')
            field = {'name': index_name, 'type': 'string', 'format': 'default'}
            fields.append(field)

        for f in mfile.findall('dwc:field', DwCStructure.ns):
            field = {}
            mterm = f.get('term')
            term = self.voc.term(mterm)
            if term is None:
                print(mterm, 'unknown Darwin Core term.')
                field['name'] = mterm.rsplit('/', 1)[1]
                field['type'] = 'string'
            else:
                field['name'] = term['name']
                field['type'] = term['type']
                field['description'] = mterm
                if term['format'] != 'default':
                    field['format'] = term['format']
                if term['constraints'] != '':
                    field['constraints'] = json.loads(term['constraints'])
            if f.get('default') is not None:
                self.has_default_values = True
            fields.append(field)
        schema['fields'] = fields
        r['schema'] = schema
        if core:
            r['primaryKey'] = self.pkey_name
        else:
            fkeys = []
            fkey = {}

            fkey['fields'] = index_name
            ref = {}
            ref['resource'] = self.corename
            ref['fields'] = self.pkey_name
            fkey['reference'] = ref
            fkeys.append(fkey)
            r['foreignKeys'] = fkeys
        return r



from FrictionlessDarwinCore import *
from pathlib import Path

import sys
import shutil
import requests
import zipfile
import tempfile
import xml.etree.ElementTree as ET


class DwCArchive:
    voc = DwCVocabulary()
    ns = {'dwc': 'http://rs.tdwg.org/dwc/text/'}

    def __init__(self, path_or_url):
        self.valid = True
        self.dwca = path_or_url
        self.meta_dir = ''
        self.metadata = None
        self.structure = None
        self.tf = None
        self.need_conversion = False
        if self.dwca.startswith('http'):
            self.path = self.download()
        else:
            self.path = self.dwca

    def download(self):
        #  download DwCArchive into temporary file
        self.tf = tempfile.NamedTemporaryFile()
        try:
            response = requests.get(self.dwca)
            self.tf.write(response.content)
        except requests.exceptions.RequestException as err:
            print(err)
            self.valid = False
        else:
            print('downloading ' + self.dwca + ' as ' + self.tf.name)
        return self.tf.name

    def infer(self):
        if zipfile.is_zipfile(self.path):
            self.load()
        else:
            self.valid = False
            print('dwca is not a zipfile')

    def load(self):
        eml = ''
        meta = ''
        zf = zipfile.ZipFile(self.path, mode='r')
        try:
            for info in zf.infolist():
                p=Path(info.filename)
                if p.name in ('eml.xml','metadata.xml'):
                    eml = zf.read(info.filename).decode()
                if p.name == 'meta.xml':
                    meta = zf.read(info.filename).decode()
                    self.meta_dir = p.parent.name
            if eml != '' and meta != '':
                self.metadata = DwCMetadata(eml)
                self.structure = DwCStructure(meta, eml, self.meta_dir)
                self.metadata.convert()
                self.structure.convert()
                self.valid = self.metadata.valid and self.structure.valid
                self.need_conversion = self.structure.has_default_values
            else:
                print('EML or Meta file missing')
                self.valid = False
        except BaseException:
            print(sys.exc_info())
            print('load zip failed')
            self.valid = False
        finally:
            zf.close()

    def _load_data(self, zf, filename, meta):
        try:
            resource = None
            for info in zf.infolist():
                p = Path(info.filename)
                if p.name == filename:
                    data = zf.read(info.filename).decode()
                    resource = DwCResource(meta, data)
        except KeyError:
            print ('Did not find %s in zip file' % p)
            self.valid = False
        except BaseException:
            print(sys.exc_info())
            print('load data from zip failed')
            self.valid = False
        return resource

    def save(self, output):
        if self.need_conversion:
            # convert all data files
            self._save_data(output)
            self._save_dwc_meta(output)
        else:
            # No Data conversion need, first copy zipfile to output
            shutil.copyfile(self.path, output)
        self._save_meta(output)

    def _save_meta(self, output):
        # Append structure and metadata
        zf = zipfile.ZipFile(output, mode='a')
        try:
             # Add a README.md file that describes the package
            zf.writestr('readme.md', self.metadata.as_markdown())
            # Add a datapackage.json
            zf.writestr('datapackage.json', self.structure.as_json())
        except IOError as e:
            print("Unable to copy file. %s" % e)
        else:
            print('saving zipfile to ' + output)
        finally:
            zf.close()

    def _save_dwc_meta(self, output):
        # Append original eml.xml , meta.xml or metadata.xml in meta_dir
        izf = zipfile.ZipFile(self.path, mode='r')
        ozf = zipfile.ZipFile(output, mode='a')
        try:
            for info in izf.infolist():
                p=Path(info.filename)
                if p.parent.name == self.meta_dir and p.suffix == '.xml' and not p.name.startswith('.'):
                    print ('copying ',info.filename, self.meta_dir)
                    ozf.writestr(p.name, izf.read(info.filename).decode())
        except BaseException:
            print(sys.exc_info())
            print('load zip failed')
            self.valid = False
        finally:
            izf.close()
            ozf.close()

    def _save_data(self, output):
        # create empty output zipfile
        izf = zipfile.ZipFile(self.path, mode='r')
        ozf = zipfile.ZipFile(output, mode='w')
        # add core and extension data as CSV files
        try:
            archive = ET.fromstring(self.structure.meta)
            core = archive.find('dwc:core', DwCArchive.ns)
            files = core.find('dwc:files', DwCArchive.ns)
            location = files.find('dwc:location', DwCArchive.ns)
            resource = self._load_data(izf, location.text, core)
            if resource is not None:
                ozf.writestr(location.text, resource.convert())
            for extension in archive.findall('dwc:extension', DwCArchive.ns):
                files = extension.find('dwc:files', DwCArchive.ns)
                location = files.find('dwc:location', DwCArchive.ns)
                resource = self._load_data(izf,location.text, extension)
                if resource is not None:
                    ozf.writestr(location.text, resource.convert())
        except IOError as e:
            print("Unable to copy file. %s" % e)
        finally:
            izf.close()
            ozf.close()

    def to_csv(self, output):
        self._save_data(output)

    def to_json(self, output):
        o = open(output, 'w')
        try:
            o.write(self.structure.as_json())
        except IOError as e:
            print("Unable to write file. %s" % e)
        else:
            print('saving json to ' + output)
        finally:
            o.close()

    def to_markdown(self, output):
        o = open(output, 'w')
        try:
            o.write(self.metadata.as_markdown())
        except IOError as e:
            print("Unable to write file. %s" % e)
        else:
            print('saving markdown to ' + output)
        finally:
            o.close()



from FrictionlessDarwinCore import DwCVocabulary
import xml.etree.ElementTree as ET
import csv
import io

class DwCResource:
    voc = DwCVocabulary()
    ns = {'dwc': 'http://rs.tdwg.org/dwc/text/'}

    def __init__(self, meta: ET, data: str):
        self.meta = meta
        self.data = data
        self.valid = True
        self.rows= []

    def convert(self):
        self.rows= []
        fields={}
        if self.meta is None:
            return
        for f in self.meta.findall('dwc:field', DwCResource.ns):
            if f.get('default') is None:
                fields[f.get('term')] = ''
            else:
                fields[f.get('term')] = f.get('default')
        datareader = csv.reader(self.data.split(self._delimiter('linesTerminatedBy')),
                                delimiter=self._delimiter('fieldsTerminatedBy'))
        header = []
        for f in fields:
            header.append(f.rsplit('/', 1)[1])
        self._append(header)
        skip = True
        for inrow in datareader:
            if skip or len(inrow) == 0:
                skip = False
            else:
                outrow=[]
                index=0
                for fname, fvalue in fields.items():
                    if fvalue == '':
                        outrow.append(inrow[index])
                        index=index+1
                    else:
                        outrow.append(fvalue)
                self._append(outrow)
        self.valid = True
        return self.as_csv()

    def as_csv(self):
        output = io.StringIO()
        quotechar=self._delimiter('fieldsEnclosedBy')
        if quotechar == '':
            quoting = csv.QUOTE_NONE
        else:
            quoting = csv.QUOTE_MINIMAL
        datawriter = csv.writer(output, lineterminator= self._delimiter('linesTerminatedBy'),
                                delimiter = self._delimiter('fieldsTerminatedBy'),
                                quotechar=quotechar,
                                quoting=quoting)
        for r in self.rows:
            datawriter.writerow(r)
        return output.getvalue()

    def _delimiter(self, delimiter_string):
        switcher = {
            "\\n": '\n',
            "\\r": '\r',
            "\\r\\n": '\r\n',
            "\\t": '\t',
        }
        ds = self.meta.get(delimiter_string)
        if ds is not None:
            return switcher.get(ds, ds)
        else:
            if delimiter_string == 'fieldsEnclosedBy':
                return '"'
            else:
                return ''

    def _append(self, row):
        self.rows.append(row)




from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='FrictionlessDarwinCore',
    version='0.2.1',
    author='André Heughebaert',
    author_email='andrejjh@gmail.com',
    license='MIT License',
    description='A tool converting Darwin Core Archive into Frictionless Data Package.',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'requests==2.22.0',
        'Click==7.0'
    ],
    entry_points='''
        [console_scripts]
        fdwca=FrictionlessDarwinCore.fdwca:cli
    ''',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.7',
)
